from strutils import isEmptyOrWhitespace
from httpcore import HttpCode

type ApiError* = object of CatchableError
  respMsg*: string
  respCode*: HttpCode


proc newApiError*(parent: ref Exception = nil, respCode: HttpCode, respMsg: string, msg = ""): ref ApiError =
  result = newException(ApiError, msg, parent)
  result.respCode = respCode
  result.respMsg = respMsg

  if not parent.isNil:
    result.trace &= parent.trace


proc raiseApiError*(respCode: HttpCode, respMsg: string, msg = "") =
  var apiError = newApiError(
    parent = nil,
    respCode = respCode,
    respMsg = respMsg,
    msg = if msg.isEmptyOrWhitespace: respMsg
          else: msg)
  raise apiError


proc raiseApiError*(respCode: HttpCode, parent: ref Exception, respMsg: string = "", msg = "") =
  var apiError = newApiError(
    parent = parent,
    respCode = respCode,
    respMsg =
      if respMsg.isEmptyOrWhitespace: parent.msg
      else: respMsg,
    msg = msg)
  raise apiError
