import std/[json, jsonutils, options, sequtils, strtabs, strutils]
import mummy, webby

import std/httpcore except HttpHeaders

import ./apierror

const CONTENT_TYPE_JSON* = "application/json"

## Response Utilities
## ------------------

type
  ApiResponse*[T] = object
    details*: Option[string]
    data*: Option[T]
    nextOffset*: Option[int]
    totalItems*: Option[int]
    nextLink*: Option[string]
    prevLink*: Option[string]


func initApiResponse*[T](
    details = none[string](),
    data: Option[T] = none[T](),
    nextOffset = none[int](),
    totalItems = none[int](),
    nextLink = none[string](),
    prevLink = none[string]()): ApiResponse[T] =
  ApiResponse[T](details: details, data: data, nextOffset: nextOffset,
    totalItems: totalItems, nextLink: nextLink, prevLink: prevLink)


func `%`*(r: ApiResponse): JsonNode =
  result = newJObject()
  if r.details.isSome: result["details"] = %r.details
  if r.data.isSome: result["data"] = %r.data
  if r.nextOffset.isSome: result["nextOffset"] = %r.nextOffset
  if r.totalItems.isSome: result["totalItems"] = %r.totalItems
  if r.nextLink.isSome: result["nextLink"] = %r.nextLink
  if r.prevLink.isSome: result["prevLink"] = %r.prevLink


func `$`*(r: ApiResponse): string = $(%r)


proc makeCorsHeaders*(
    allowedMethods: seq[string],
    allowedOrigins: seq[string],
    reqOrigin = none[string]()): HttpHeaders =

  result =
    if reqOrigin.isSome and allowedOrigins.contains(reqOrigin.get):
      @{
        "Access-Control-Allow-Origin": reqOrigin.get,
        "Access-Control-Allow-Credentials": "true",
        "Access-Control-Allow-Methods": allowedMethods.join(", "),
        "Access-Control-Allow-Headers": "DNT,User-Agent,X-Requested-With,If-Modified-Since,Cache-Control,Content-Type,Range,Authorization,X-CSRF-TOKEN"
      }
    else:
      if reqOrigin.isSome:
        @{"X-Invalid-Origin-Details": "Unrecognized origin '" & reqOrigin.get & "'."}
      else:
        @{"X-Invalid-Origin-Details": "Missing Origin."}


proc makeCorsHeaders*(
    allowedMethods: seq[HttpMethod],
    allowedOrigins: seq[string],
    reqOrigin = none[string]()): HttpHeaders =
 makeCorsHeaders(allowedMethods.mapIt($it), allowedOrigins, reqOrigin )


func origin*(req: Request): Option[string] =
  if req.headers.contains("Origin"): some(req.headers["Origin"])
  else: none[string]()


proc respondWithRawJson*(
    req: Request,
    body: JsonNode,
    code = Http200,
    allowedOrigins = newSeq[string](),
    headersToSend: HttpHeaders = @{:}) =
  ## Immediately send a JSON response and stop processing the request.

  var headers =
    headersToSend &
    makeCorsHeaders(@[req.httpMethod], allowedOrigins, req.origin) &
    @[("Content-Type", CONTENT_TYPE_JSON),
      ("Cache-Control", "no-cache")]

  req.respond(code.ord, headers, $body)


proc respond*[T](
    req: Request,
    resp: ApiResponse[T],
    code = Http200,
    allowedOrigins = newSeq[string](),
    headersToSend: HttpHeaders = @{:}) =
  req.respondWithRawJson(%resp, code, allowedOrigins, headersToSend)


proc respondWithData*[T](
    req: Request,
    data: T,
    code = Http200,
    allowedOrigins = newSeq[string](),
    headersToSend: HttpHeaders = @{:}) =
  req.respond(initApiResponse[T](data = some(data)),
    code, allowedOrigins, headersToSend)


proc respondToOptions*(
    req: Request,
    allowedMethods: seq[HttpMethod],
    allowedOrigins: seq[string]) =

  req.respond(
    Http200.ord,
    makeCorsHeaders(allowedMethods, allowedOrigins, req.origin),
    "")
